@@ -26,12 +26,19 @@
 #include <grub/scsi.h>
 #include <grub/scsicmd.h>
 #include <grub/time.h>
+#include <grub/i18n.h>
 
 GRUB_MOD_LICENSE ("GPLv3+");
 
 
 static grub_scsi_dev_t grub_scsi_dev_list;
 
+const char grub_scsi_names[GRUB_SCSI_NUM_SUBSYSTEMS][5] = {
+  [GRUB_SCSI_SUBSYSTEM_USBMS] = "usb",
+  [GRUB_SCSI_SUBSYSTEM_PATA] = "ata",
+  [GRUB_SCSI_SUBSYSTEM_AHCI] = "ahci"
+};
+
 void
 grub_scsi_dev_register (grub_scsi_dev_t dev)
 {
@@ -147,14 +154,14 @@ grub_scsi_inquiry (grub_scsi_t scsi)
 
 /* Read the capacity and block size of SCSI.  */
 static grub_err_t
-grub_scsi_read_capacity (grub_scsi_t scsi)
+grub_scsi_read_capacity10 (grub_scsi_t scsi)
 {
-  struct grub_scsi_read_capacity rc;
-  struct grub_scsi_read_capacity_data rcd;
+  struct grub_scsi_read_capacity10 rc;
+  struct grub_scsi_read_capacity10_data rcd;
   grub_err_t err;
   grub_err_t err_sense;
 
-  rc.opcode = grub_scsi_cmd_read_capacity;
+  rc.opcode = grub_scsi_cmd_read_capacity10;
   rc.lun = scsi->lun << GRUB_SCSI_LUN_SHIFT;
   rc.logical_block_addr = 0;
   rc.reserved1 = 0;
@@ -176,7 +183,42 @@ grub_scsi_read_capacity (grub_scsi_t scsi)
   if (err)
     return err;
 
-  scsi->size = grub_be_to_cpu32 (rcd.size);
+  scsi->last_block = grub_be_to_cpu32 (rcd.last_block);
+  scsi->blocksize = grub_be_to_cpu32 (rcd.blocksize);
+
+  return GRUB_ERR_NONE;
+}
+
+/* Read the capacity and block size of SCSI.  */
+static grub_err_t
+grub_scsi_read_capacity16 (grub_scsi_t scsi)
+{
+  struct grub_scsi_read_capacity16 rc;
+  struct grub_scsi_read_capacity16_data rcd;
+  grub_err_t err;
+  grub_err_t err_sense;
+
+  rc.opcode = grub_scsi_cmd_read_capacity16;
+  rc.lun = (scsi->lun << GRUB_SCSI_LUN_SHIFT) | 0x10;
+  rc.logical_block_addr = 0;
+  rc.alloc_len = grub_cpu_to_be32 (sizeof (rcd));
+  rc.PMI = 0;
+  rc.control = 0;
+	
+  err = scsi->dev->read (scsi, sizeof (rc), (char *) &rc,
+			 sizeof (rcd), (char *) &rcd);
+
+  /* Each SCSI command should be followed by Request Sense.
+     If not so, many devices STALLs or definitely freezes. */
+  err_sense = grub_scsi_request_sense (scsi);
+  if (err_sense != GRUB_ERR_NONE)
+  	grub_errno = err;
+/* err_sense is ignored for now and Request Sense Data also... */
+
+  if (err)
+    return err;
+
+  scsi->last_block = grub_be_to_cpu64 (rcd.last_block);
   scsi->blocksize = grub_be_to_cpu32 (rcd.blocksize);
 
   return GRUB_ERR_NONE;
@@ -247,12 +289,43 @@ grub_scsi_read12 (grub_disk_t disk, grub_disk_addr_t sector,
   return err;
 }
 
-#if 0
+/* Send a SCSI request for DISK: read SIZE sectors starting with
+   sector SECTOR to BUF.  */
+static grub_err_t
+grub_scsi_read16 (grub_disk_t disk, grub_disk_addr_t sector,
+		  grub_size_t size, char *buf)
+{
+  grub_scsi_t scsi;
+  struct grub_scsi_read16 rd;
+  grub_err_t err;
+  grub_err_t err_sense;
+
+  scsi = disk->data;
+
+  rd.opcode = grub_scsi_cmd_read16;
+  rd.lun = scsi->lun << GRUB_SCSI_LUN_SHIFT;
+  rd.lba = grub_cpu_to_be64 (sector);
+  rd.size = grub_cpu_to_be32 (size);
+  rd.reserved = 0;
+  rd.control = 0;
+
+  err = scsi->dev->read (scsi, sizeof (rd), (char *) &rd, size * scsi->blocksize, buf);
+
+  /* Each SCSI command should be followed by Request Sense.
+     If not so, many devices STALLs or definitely freezes. */
+  err_sense = grub_scsi_request_sense (scsi);
+  if (err_sense != GRUB_ERR_NONE)
+  	grub_errno = err;
+  /* err_sense is ignored for now and Request Sense Data also... */
+
+  return err;
+}
+
 /* Send a SCSI request for DISK: write the data stored in BUF to SIZE
    sectors starting with SECTOR.  */
 static grub_err_t
 grub_scsi_write10 (grub_disk_t disk, grub_disk_addr_t sector,
-		   grub_size_t size, char *buf)
+		   grub_size_t size, const char *buf)
 {
   grub_scsi_t scsi;
   struct grub_scsi_write10 wr;
@@ -281,6 +354,8 @@ grub_scsi_write10 (grub_disk_t disk, grub_disk_addr_t sector,
   return err;
 }
 
+#if 0
+
 /* Send a SCSI request for DISK: write the data stored in BUF to SIZE
    sectors starting with SECTOR.  */
 static grub_err_t
@@ -314,15 +389,49 @@ grub_scsi_write12 (grub_disk_t disk, grub_disk_addr_t sector,
 }
 #endif
 
+/* Send a SCSI request for DISK: write the data stored in BUF to SIZE
+   sectors starting with SECTOR.  */
+static grub_err_t
+grub_scsi_write16 (grub_disk_t disk, grub_disk_addr_t sector,
+		   grub_size_t size, const char *buf)
+{
+  grub_scsi_t scsi;
+  struct grub_scsi_write16 wr;
+  grub_err_t err;
+  grub_err_t err_sense;
+
+  scsi = disk->data;
+
+  wr.opcode = grub_scsi_cmd_write16;
+  wr.lun = scsi->lun << GRUB_SCSI_LUN_SHIFT;
+  wr.lba = grub_cpu_to_be64 (sector);
+  wr.size = grub_cpu_to_be32 (size);
+  wr.reserved = 0;
+  wr.control = 0;
+
+  err = scsi->dev->write (scsi, sizeof (wr), (char *) &wr, size * scsi->blocksize, buf);
+
+  /* Each SCSI command should be followed by Request Sense.
+     If not so, many devices STALLs or definitely freezes. */
+  err_sense = grub_scsi_request_sense (scsi);
+  if (err_sense != GRUB_ERR_NONE)
+  	grub_errno = err;
+  /* err_sense is ignored for now and Request Sense Data also... */
+
+  return err;
+}
+
+
 
 static int
-grub_scsi_iterate (int (*hook) (const char *name))
+grub_scsi_iterate (int (*hook) (const char *name),
+		   grub_disk_pull_t pull)
 {
   grub_scsi_dev_t p;
 
-  auto int scsi_iterate (int bus, int luns);
+  auto int NESTED_FUNC_ATTR scsi_iterate (int id, int bus, int luns);
 
-  int scsi_iterate (int bus, int luns)
+  int NESTED_FUNC_ATTR scsi_iterate (int id, int bus, int luns)
     {
       int i;
 
@@ -331,7 +440,7 @@ grub_scsi_iterate (int (*hook) (const char *name))
 	{
 	  char *sname;
 	  int ret;
-	  sname = grub_xasprintf ("%s%d", p->name, bus);
+	  sname = grub_xasprintf ("%s%d", grub_scsi_names[id], bus);
 	  if (!sname)
 	    return 1;
 	  ret = hook (sname);
@@ -345,7 +454,7 @@ grub_scsi_iterate (int (*hook) (const char *name))
 	{
 	  char *sname;
 	  int ret;
-	  sname = grub_xasprintf ("%s%d%c", p->name, bus, 'a' + i);
+	  sname = grub_xasprintf ("%s%d%c", grub_scsi_names[id], bus, 'a' + i);
 	  if (!sname)
 	    return 1;
 	  ret = hook (sname);
@@ -357,7 +466,7 @@ grub_scsi_iterate (int (*hook) (const char *name))
     }
 
   for (p = grub_scsi_dev_list; p; p = p->next)
-    if (p->iterate && (p->iterate) (scsi_iterate))
+    if (p->iterate && (p->iterate) (scsi_iterate, pull))
       return 1;
 
   return 0;
@@ -372,6 +481,7 @@ grub_scsi_open (const char *name, grub_disk_t disk)
   int lun, bus;
   grub_uint64_t maxtime;
   const char *nameend;
+  unsigned id;
 
   nameend = name + grub_strlen (name) - 1;
   /* Try to detect a LUN ('a'-'z'), otherwise just use the first
@@ -396,15 +506,25 @@ grub_scsi_open (const char *name, grub_disk_t disk)
   if (! scsi)
     return grub_errno;
 
+  for (id = 0; id < ARRAY_SIZE (grub_scsi_names); id++)
+    if (grub_strncmp (grub_scsi_names[id], name, nameend - name) == 0)
+      break;
+
+  if (id == ARRAY_SIZE (grub_scsi_names))
+    {
+      grub_free (scsi);
+      return grub_error (GRUB_ERR_UNKNOWN_DEVICE, "not a SCSI disk");
+    }
+
   for (p = grub_scsi_dev_list; p; p = p->next)
     {
-      if (grub_strncmp (p->name, name, nameend - name) != 0)
-	continue;
+      if (p->open (id, bus, scsi))
+	{
+	  grub_errno = GRUB_ERR_NONE;
+	  continue;
+	}
 
-      if (p->open (bus, scsi))
-	continue;
-
-      disk->id = grub_make_scsi_id (p->id, bus, lun);
+      disk->id = grub_make_scsi_id (id, bus, lun);
       disk->data = scsi;
       scsi->dev = p;
       scsi->lun = lun;
@@ -457,30 +577,45 @@ grub_scsi_open (const char *name, grub_disk_t disk)
       grub_errno = GRUB_ERR_NONE;
 
       /* Read capacity of media */
-      err = grub_scsi_read_capacity (scsi);
+      err = grub_scsi_read_capacity10 (scsi);
       if (err)
 	{
 	  grub_free (scsi);
-	  grub_dprintf ("scsi", "READ CAPACITY failed\n");
+	  grub_dprintf ("scsi", "READ CAPACITY10 failed\n");
 	  return err;
 	}
 
-      /* SCSI blocks can be something else than 512, although GRUB
-	 wants 512 byte blocks.  */
-      disk->total_sectors = ((grub_uint64_t)scsi->size
-                             * (grub_uint64_t)scsi->blocksize)
-			    >> GRUB_DISK_SECTOR_BITS;
+      if (scsi->last_block == 0xffffffff)
+	{
+	  err = grub_scsi_read_capacity16 (scsi);
+	  if (err)
+	    {
+	      grub_free (scsi);
+	      grub_dprintf ("scsi", "READ CAPACITY16 failed\n");
+	      return err;
+	    }
+	}
 
-      grub_dprintf ("scsi", "blocks=%u, blocksize=%u\n",
-		    scsi->size, scsi->blocksize);
-      grub_dprintf ("scsi", "Disk total 512 sectors = %llu\n",
+      disk->total_sectors = scsi->last_block + 1;
+      if (scsi->blocksize & (scsi->blocksize - 1) || !scsi->blocksize)
+	{
+	  grub_free (scsi);
+	  return grub_error (GRUB_ERR_IO, "invalid sector size %d",
+			     scsi->blocksize);
+	}
+      for (disk->log_sector_size = 0;
+	   (1U << disk->log_sector_size) < scsi->blocksize;
+	   disk->log_sector_size++);
+
+      grub_dprintf ("scsi", "last_block=%" PRIuGRUB_UINT64_T ", blocksize=%u\n",
+		    scsi->last_block, scsi->blocksize);
+      grub_dprintf ("scsi", "Disk total sectors = %llu\n",
 		    (unsigned long long) disk->total_sectors);
 
       return GRUB_ERR_NONE;
     }
 
   grub_free (scsi);
-
   return grub_error (GRUB_ERR_UNKNOWN_DEVICE, "not a SCSI disk");
 }
 
@@ -503,36 +638,42 @@ grub_scsi_read (grub_disk_t disk, grub_disk_addr_t sector,
 
   scsi = disk->data;
 
-  /* SCSI sectors are variable in size.  GRUB uses 512 byte
-     sectors.  */
-  if (scsi->blocksize != GRUB_DISK_SECTOR_SIZE)
+  while (size)
     {
-      unsigned spb = scsi->blocksize >> GRUB_DISK_SECTOR_BITS;
-      if (spb == 0 || (scsi->blocksize & (GRUB_DISK_SECTOR_SIZE - 1)) != 0)
-	return grub_error (GRUB_ERR_NOT_IMPLEMENTED_YET,
-			   "unsupported SCSI block size");
+      /* PATA doesn't support more than 32K reads.
+	 Not sure about AHCI and USB. If it's confirmed that either of
+	 them can do bigger reads reliably this value can be moved to 'scsi'
+	 structure.  */
+      grub_size_t len = 32768 >> disk->log_sector_size;
+      grub_err_t err;
+      if (len > size)
+	len = size;
+      /* Depending on the type, select a read function.  */
+      switch (scsi->devtype)
+	{
+	case grub_scsi_devtype_direct:
+	  if (sector >> 32)
+	    err = grub_scsi_read16 (disk, sector, len, buf);
+	  else
+	    err = grub_scsi_read10 (disk, sector, len, buf);
+	  if (err)
+	    return err;
+	  break;
 
-      grub_uint32_t sector_mod = 0;
-      sector = grub_divmod64 (sector, spb, &sector_mod);
-
-      if (! (sector_mod == 0 && size % spb == 0))
-	return grub_error (GRUB_ERR_NOT_IMPLEMENTED_YET,
-			   "unaligned SCSI read not supported");
-
-      size /= spb;
+	case grub_scsi_devtype_cdrom:
+	  if (sector >> 32)
+	    err = grub_scsi_read16 (disk, sector, len, buf);
+	  else
+	    err = grub_scsi_read12 (disk, sector, len, buf);
+	  if (err)
+	    return err;
+	  break;
+	}
+      size -= len;
+      sector += len;
+      buf += len << disk->log_sector_size;
     }
 
-  /* Depending on the type, select a read function.  */
-  switch (scsi->devtype)
-    {
-    case grub_scsi_devtype_direct:
-      return grub_scsi_read10 (disk, sector, size, buf);
-
-    case grub_scsi_devtype_cdrom:
-      return grub_scsi_read12 (disk, sector, size, buf);
-    }
-
-  /* XXX: Never reached.  */
   return GRUB_ERR_NONE;
 
 #if 0 /* Workaround - it works - but very slowly, from some reason
@@ -573,13 +714,41 @@ grub_scsi_write (grub_disk_t disk __attribute((unused)),
 		 grub_size_t size __attribute((unused)),
 		 const char *buf __attribute((unused)))
 {
-#if 0
-  /* XXX: Not tested yet!  */
+  grub_scsi_t scsi;
 
-  /* XXX: This should depend on the device type?  */
-  return grub_scsi_write10 (disk, sector, size, buf);
-#endif
-  return GRUB_ERR_NOT_IMPLEMENTED_YET;
+  scsi = disk->data;
+
+  if (scsi->devtype == grub_scsi_devtype_cdrom)
+    return grub_error (GRUB_ERR_IO, N_("cannot write to CD-ROM"));
+
+  while (size)
+    {
+      /* PATA doesn't support more than 32K reads.
+	 Not sure about AHCI and USB. If it's confirmed that either of
+	 them can do bigger reads reliably this value can be moved to 'scsi'
+	 structure.  */
+      grub_size_t len = 32768 >> disk->log_sector_size;
+      grub_err_t err;
+      if (len > size)
+	len = size;
+      /* Depending on the type, select a read function.  */
+      switch (scsi->devtype)
+	{
+	case grub_scsi_devtype_direct:
+	  if (sector >> 32)
+	    err = grub_scsi_write16 (disk, sector, len, buf);
+	  else
+	    err = grub_scsi_write10 (disk, sector, len, buf);
+	  if (err)
+	    return err;
+	  break;
+	}
+      size -= len;
+      sector += len;
+      buf += len << disk->log_sector_size;
+    }
+
+  return GRUB_ERR_NONE;
 }
 
 
