@@ -36,7 +36,6 @@ GRUB_MOD_LICENSE ("GPLv3+");
 
 #include "types.h"
 #include "g10lib.h"
-#include "memory.h"
 #include "cipher.h"
 
 #include "bithelp.h"
@@ -1404,6 +1403,9 @@ gcry_md_spec_t _gcry_digest_spec_whirlpool =
     whirlpool_init, whirlpool_write, whirlpool_final, whirlpool_read,
     sizeof (whirlpool_context_t)
     ,
+#ifdef GRUB_UTIL
+    .modname = "gcry_whirlpool",
+#endif
     .blocksize = 64
   };
 
