@@ -11,6 +11,10 @@ typedef struct grub_macho_thread32 grub_macho_thread_t;
 #define cmdsizeXX cmdsize32
 #define cmdsXX cmds32
 #define endXX end32
+#define uncompressedXX uncompressed32
+#define compressedXX compressed32
+#define uncompressed_sizeXX uncompressed_size32
+#define compressed_sizeXX compressed_size32
 #define XX "32"
 #define GRUB_MACHO_MAGIC GRUB_MACHO_MAGIC32
 #define GRUB_MACHO_CMD_SEGMENT GRUB_MACHO_CMD_SEGMENT32
