@@ -11,6 +11,10 @@ typedef struct grub_macho_thread64 grub_macho_thread_t;
 #define cmdsizeXX cmdsize64
 #define cmdsXX cmds64
 #define endXX end64
+#define uncompressedXX uncompressed64
+#define compressedXX compressed64
+#define uncompressed_sizeXX uncompressed_size64
+#define compressed_sizeXX compressed_size64
 #define XX "64"
 #define GRUB_MACHO_MAGIC GRUB_MACHO_MAGIC64
 #define GRUB_MACHO_CMD_SEGMENT GRUB_MACHO_CMD_SEGMENT64
