@@ -42,13 +42,6 @@
   Most underline diacritics aren't displayed in gfxterm
  */
 
-/* Convert a (possibly null-terminated) UTF-8 string of at most SRCSIZE
-   bytes (if SRCSIZE is -1, it is ignored) in length to a UTF-16 string.
-   Return the number of characters converted. DEST must be able to hold
-   at least DESTSIZE characters. If an invalid sequence is found, return -1.
-   If SRCEND is not NULL, then *SRCEND is set to the next byte after the
-   last byte used in SRC.  */
-
 #include <grub/charset.h>
 #include <grub/mm.h>
 #include <grub/misc.h>
@@ -60,142 +53,88 @@
 #include "widthspec.h"
 #endif
 
+/* Returns -2 if not enough space, -1 on invalid character.  */
 grub_ssize_t
-grub_utf8_to_utf16 (grub_uint16_t *dest, grub_size_t destsize,
-		    const grub_uint8_t *src, grub_size_t srcsize,
-		    const grub_uint8_t **srcend)
+grub_encode_utf8_character (grub_uint8_t *dest, grub_uint8_t *destend,
+			    grub_uint32_t code)
 {
-  grub_uint16_t *p = dest;
-  int count = 0;
-  grub_uint32_t code = 0;
-
-  if (srcend)
-    *srcend = src;
-
-  while (srcsize && destsize)
+  if (dest >= destend)
+    return -2;
+  if (code <= 0x007F)
     {
-      grub_uint32_t c = *src++;
-      if (srcsize != (grub_size_t)-1)
-	srcsize--;
-      if (count)
-	{
-	  if ((c & GRUB_UINT8_2_LEADINGBITS) != GRUB_UINT8_1_LEADINGBIT)
-	    {
-	      /* invalid */
-	      return -1;
-	    }
-	  else
-	    {
-	      code <<= 6;
-	      code |= (c & GRUB_UINT8_6_TRAILINGBITS);
-	      count--;
-	    }
-	}
-      else
-	{
-	  if (c == 0)
-	    break;
-
-	  if ((c & GRUB_UINT8_1_LEADINGBIT) == 0)
-	    code = c;
-	  else if ((c & GRUB_UINT8_3_LEADINGBITS) == GRUB_UINT8_2_LEADINGBITS)
-	    {
-	      count = 1;
-	      code = c & GRUB_UINT8_5_TRAILINGBITS;
-	    }
-	  else if ((c & GRUB_UINT8_4_LEADINGBITS) == GRUB_UINT8_3_LEADINGBITS)
-	    {
-	      count = 2;
-	      code = c & GRUB_UINT8_4_TRAILINGBITS;
-	    }
-	  else if ((c & GRUB_UINT8_5_LEADINGBITS) == GRUB_UINT8_4_LEADINGBITS)
-	    {
-	      count = 3;
-	      code = c & GRUB_UINT8_3_TRAILINGBITS;
-	    }
-	  else
-	    return -1;
-	}
-
-      if (count == 0)
-	{
-	  if (destsize < 2 && code >= GRUB_UCS2_LIMIT)
-	    break;
-	  if (code >= GRUB_UCS2_LIMIT)
-	    {
-	      *p++ = GRUB_UTF16_UPPER_SURROGATE (code);
-	      *p++ = GRUB_UTF16_LOWER_SURROGATE (code);
-	      destsize -= 2;
-	    }
-	  else
-	    {
-	      *p++ = code;
-	      destsize--;
-	    }
-	}
+      *dest++ = code;
+      return 1;
     }
+  if (code <= 0x07FF)
+    {
+      if (dest + 1 >= destend)
+	return -2;
+      *dest++ = (code >> 6) | 0xC0;
+      *dest++ = (code & 0x3F) | 0x80;
+      return 2;
+    }
+  if ((code >= 0xDC00 && code <= 0xDFFF)
+      || (code >= 0xD800 && code <= 0xDBFF))
+    {
+      /* No surrogates in UCS-4... */
+      return -1;
+    }
+  if (code < 0x10000)
+    {
+      if (dest + 2 >= destend)
+	return -2;
+      *dest++ = (code >> 12) | 0xE0;
+      *dest++ = ((code >> 6) & 0x3F) | 0x80;
+      *dest++ = (code & 0x3F) | 0x80;
+      return 3;
+    }
+  {
+    if (dest + 3 >= destend)
+      return -2;
+    *dest++ = (code >> 18) | 0xF0;
+    *dest++ = ((code >> 12) & 0x3F) | 0x80;
+    *dest++ = ((code >> 6) & 0x3F) | 0x80;
+    *dest++ = (code & 0x3F) | 0x80;
+    return 4;
+  }
 
-  if (srcend)
-    *srcend = src;
-  return p - dest;
 }
 
 /* Convert UCS-4 to UTF-8.  */
-void
-grub_ucs4_to_utf8 (grub_uint32_t *src, grub_size_t size,
+grub_size_t
+grub_ucs4_to_utf8 (const grub_uint32_t *src, grub_size_t size,
 		   grub_uint8_t *dest, grub_size_t destsize)
 {
   /* Keep last char for \0.  */
   grub_uint8_t *destend = dest + destsize - 1;
+  grub_uint8_t *dest0 = dest;
 
   while (size-- && dest < destend)
     {
       grub_uint32_t code = *src++;
-
-      if (code <= 0x007F)
-	*dest++ = code;
-      else if (code <= 0x07FF)
+      grub_ssize_t s;
+      s = grub_encode_utf8_character (dest, destend, code);
+      if (s == -2)
+	break;
+      if (s == -1)
 	{
-	  if (dest + 1 >= destend)
-	    break;
-	  *dest++ = (code >> 6) | 0xC0;
-	  *dest++ = (code & 0x3F) | 0x80;
-	}
-      else if ((code >= 0xDC00 && code <= 0xDFFF)
-	       || (code >= 0xD800 && code <= 0xDBFF))
-	{
-	  /* No surrogates in UCS-4... */
 	  *dest++ = '?';
+	  continue;
 	}
-      else if (code < 0x10000)
-	{
-	  if (dest + 2 >= destend)
-	    break;
-	  *dest++ = (code >> 12) | 0xE0;
-	  *dest++ = ((code >> 6) & 0x3F) | 0x80;
-	  *dest++ = (code & 0x3F) | 0x80;
-	}
-      else
-	{
-	  if (dest + 3 >= destend)
-	    break;
-	  *dest++ = (code >> 18) | 0xF0;
-	  *dest++ = ((code >> 12) & 0x3F) | 0x80;
-	  *dest++ = ((code >> 6) & 0x3F) | 0x80;
-	  *dest++ = (code & 0x3F) | 0x80;
-	}
+      dest += s;
     }
   *dest = 0;
+  return dest - dest0;
 }
 
-/* Convert UCS-4 to UTF-8.  */
-char *
-grub_ucs4_to_utf8_alloc (grub_uint32_t *src, grub_size_t size)
+/* Returns the number of bytes the string src would occupy is converted
+   to UTF-8, excluding trailing \0.  */
+grub_size_t
+grub_get_num_of_utf8_bytes (const grub_uint32_t *src, grub_size_t size)
 {
   grub_size_t remaining;
-  grub_uint32_t *ptr;
+  const grub_uint32_t *ptr;
   grub_size_t cnt = 0;
-  grub_uint8_t *ret;
 
   remaining = size;
   ptr = src;
@@ -216,7 +155,15 @@ grub_ucs4_to_utf8_alloc (grub_uint32_t *src, grub_size_t size)
       else
 	cnt += 4;
     }
-  cnt++;
+  return cnt;
+}
+
+/* Convert UCS-4 to UTF-8.  */
+char *
+grub_ucs4_to_utf8_alloc (const grub_uint32_t *src, grub_size_t size)
+{
+  grub_uint8_t *ret;
+  grub_size_t cnt = grub_get_num_of_utf8_bytes (src, size) + 1;
 
   ret = grub_malloc (cnt);
   if (!ret)
@@ -230,61 +177,29 @@ grub_ucs4_to_utf8_alloc (grub_uint32_t *src, grub_size_t size)
 int
 grub_is_valid_utf8 (const grub_uint8_t *src, grub_size_t srcsize)
 {
-  grub_uint32_t code = 0;
   int count = 0;
+  grub_uint32_t code = 0;
 
   while (srcsize)
     {
-      grub_uint32_t c = *src++;
       if (srcsize != (grub_size_t)-1)
 	srcsize--;
-      if (count)
-	{
-	  if ((c & 0xc0) != 0x80)
-	    {
-	      /* invalid */
-	      return 0;
-	    }
-	  else
-	    {
-	      code <<= 6;
-	      code |= (c & 0x3f);
-	      count--;
-	    }
-	}
-      else
-	{
-	  if (c == 0)
-	    break;
-
-	  if ((c & 0x80) == 0x00)
-	    code = c;
-	  else if ((c & 0xe0) == 0xc0)
-	    {
-	      count = 1;
-	      code = c & 0x1f;
-	    }
-	  else if ((c & 0xf0) == 0xe0)
-	    {
-	      count = 2;
-	      code = c & 0x0f;
-	    }
-	  else if ((c & 0xf8) == 0xf0)
-	    {
-	      count = 3;
-	      code = c & 0x07;
-	    }
-	  else
-	    return 0;
-	}
+      if (!grub_utf8_process (*src++, &code, &count))
+	return 0;
+      if (count != 0)
+	continue;
+      if (code == 0)
+	return 1;
+      if (code > GRUB_UNICODE_LAST_VALID)
+	return 0;
     }
 
   return 1;
 }
 
-int
+grub_ssize_t
 grub_utf8_to_ucs4_alloc (const char *msg, grub_uint32_t **unicode_msg,
-			grub_uint32_t **last_position)
+			 grub_uint32_t **last_position)
 {
   grub_size_t msg_len = grub_strlen (msg);
 
@@ -322,63 +237,23 @@ grub_utf8_to_ucs4 (grub_uint32_t *dest, grub_size_t destsize,
 
   while (srcsize && destsize)
     {
-      grub_uint32_t c = *src++;
+      int was_count = count;
       if (srcsize != (grub_size_t)-1)
 	srcsize--;
-      if (count)
+      if (!grub_utf8_process (*src++, &code, &count))
 	{
-	  if ((c & 0xc0) != 0x80)
-	    {
-	      /* invalid */
-	      code = '?';
-	      /* Character c may be valid, don't eat it.  */
-	      src--;
-	      if (srcsize != (grub_size_t)-1)
-		srcsize++;
-	      count = 0;
-	    }
-	  else
-	    {
-	      code <<= 6;
-	      code |= (c & 0x3f);
-	      count--;
-	    }
-	}
-      else
-	{
-	  if (c == 0)
-	    break;
-
-	  if ((c & 0x80) == 0x00)
-	    code = c;
-	  else if ((c & 0xe0) == 0xc0)
-	    {
-	      count = 1;
-	      code = c & 0x1f;
-	    }
-	  else if ((c & 0xf0) == 0xe0)
-	    {
-	      count = 2;
-	      code = c & 0x0f;
-	    }
-	  else if ((c & 0xf8) == 0xf0)
-	    {
-	      count = 3;
-	      code = c & 0x07;
-	    }
-	  else
-	    {
-	      /* invalid */
-	      code = '?';
-	      count = 0;
-	    }
-	}
-
-      if (count == 0)
-	{
-	  *p++ = code;
-	  destsize--;
+	  code = '?';
+	  count = 0;
+	  /* Character c may be valid, don't eat it.  */
+	  if (was_count)
+	    src--;
 	}
+      if (count != 0)
+	continue;
+      if (code == 0)
+	break;
+      *p++ = code;
+      destsize--;
     }
 
   if (srcend)
@@ -400,9 +275,9 @@ unpack_join (void)
       grub_errno = GRUB_ERR_NONE;
       return;
     }
-  for (cur = grub_unicode_compact; cur->end; cur++)
-    for (i = cur->start; i <= cur->end
-	     && i < GRUB_UNICODE_MAX_CACHED_CHAR; i++)
+  for (cur = grub_unicode_compact; cur->len; cur++)
+    for (i = cur->start; i < cur->start + (unsigned) cur->len
+	   && i < GRUB_UNICODE_MAX_CACHED_CHAR; i++)
       join_types[i] = cur->join_type;
 }
 
@@ -420,9 +295,9 @@ unpack_bidi (void)
       grub_errno = GRUB_ERR_NONE;
       return;
     }
-  for (cur = grub_unicode_compact; cur->end; cur++)
-    for (i = cur->start; i <= cur->end
-	     && i < GRUB_UNICODE_MAX_CACHED_CHAR; i++)
+  for (cur = grub_unicode_compact; cur->len; cur++)
+    for (i = cur->start; i < cur->start + (unsigned) cur->len
+	   && i < GRUB_UNICODE_MAX_CACHED_CHAR; i++)
       if (cur->bidi_mirror)
 	bidi_types[i] = cur->bidi_type | 0x80;
       else
@@ -440,8 +315,8 @@ get_bidi_type (grub_uint32_t c)
   if (bidi_types && c < GRUB_UNICODE_MAX_CACHED_CHAR)
     return bidi_types[c] & 0x7f;
 
-  for (cur = grub_unicode_compact; cur->end; cur++)
-    if (cur->start <= c && c <= cur->end)
+  for (cur = grub_unicode_compact; cur->len; cur++)
+    if (cur->start <= c && c < cur->start + (unsigned) cur->len)
       return cur->bidi_type;
 
   return GRUB_BIDI_TYPE_L;
@@ -458,8 +333,8 @@ get_join_type (grub_uint32_t c)
   if (join_types && c < GRUB_UNICODE_MAX_CACHED_CHAR)
     return join_types[c];
 
-  for (cur = grub_unicode_compact; cur->end; cur++)
-    if (cur->start <= c && c <= cur->end)
+  for (cur = grub_unicode_compact; cur->len; cur++)
+    if (cur->start <= c && c < cur->start + (unsigned) cur->len)
       return cur->join_type;
 
   return GRUB_JOIN_TYPE_NONJOINING;
@@ -476,8 +351,8 @@ is_mirrored (grub_uint32_t c)
   if (bidi_types && c < GRUB_UNICODE_MAX_CACHED_CHAR)
     return !!(bidi_types[c] & 0x80);
 
-  for (cur = grub_unicode_compact; cur->end; cur++)
-    if (cur->start <= c && c <= cur->end)
+  for (cur = grub_unicode_compact; cur->len; cur++)
+    if (cur->start <= c && c < cur->start + (unsigned) cur->len)
       return cur->bidi_mirror;
 
   return 0;
@@ -494,8 +369,8 @@ grub_unicode_get_comb_type (grub_uint32_t c)
       unsigned i;
       comb_types = grub_zalloc (GRUB_UNICODE_MAX_CACHED_CHAR);
       if (comb_types)
-	for (cur = grub_unicode_compact; cur->end; cur++)
-	  for (i = cur->start; i <= cur->end
+	for (cur = grub_unicode_compact; cur->len; cur++)
+	  for (i = cur->start; i < cur->start + (unsigned) cur->len
 		 && i < GRUB_UNICODE_MAX_CACHED_CHAR; i++)
 	    comb_types[i] = cur->comb_type;
       else
@@ -505,8 +380,8 @@ grub_unicode_get_comb_type (grub_uint32_t c)
   if (comb_types && c < GRUB_UNICODE_MAX_CACHED_CHAR)
     return comb_types[c];
 
-  for (cur = grub_unicode_compact; cur->end; cur++)
-    if (cur->start <= c && c <= cur->end)
+  for (cur = grub_unicode_compact; cur->len; cur++)
+    if (cur->start <= c && c < cur->start + (unsigned) cur->len)
       return cur->comb_type;
 
   return GRUB_UNICODE_COMB_NONE;
@@ -553,6 +428,16 @@ grub_unicode_aglomerate_comb (const grub_uint32_t *in, grub_size_t inlen,
 
   grub_memset (out, 0, sizeof (*out));
 
+  if (inlen && grub_iscntrl (*in))
+    {
+      out->base = *in;
+      out->variant = 0;
+      out->attributes = 0;
+      out->ncomb = 0;
+      out->estimated_width = 1;
+      return 1;
+    }
+
   for (ptr = in; ptr < in + inlen; ptr++)
     {
       /* Variation selectors >= 17 are outside of BMP and SMP. 
@@ -564,7 +449,6 @@ grub_unicode_aglomerate_comb (const grub_uint32_t *in, grub_size_t inlen,
 	  if (haveout)
 	    out->variant = *ptr - GRUB_UNICODE_VARIATION_SELECTOR_1 + 1;
 	  continue;
-
 	}
       if (*ptr >= GRUB_UNICODE_VARIATION_SELECTOR_17
 	  && *ptr <= GRUB_UNICODE_VARIATION_SELECTOR_256)
@@ -588,25 +472,42 @@ grub_unicode_aglomerate_comb (const grub_uint32_t *in, grub_size_t inlen,
 	      || comb_type == GRUB_UNICODE_COMB_ME
 	      || comb_type == GRUB_UNICODE_COMB_MN)
 	    last_comb_pointer = out->ncomb;
-	  n = grub_realloc (out->combining,
-			    sizeof (n[0]) * (out->ncomb + 1));
-	  if (!n)
+
+	  if (out->ncomb + 1 <= (int) ARRAY_SIZE (out->combining_inline))
+	    n = out->combining_inline;
+	  else if (out->ncomb > (int) ARRAY_SIZE (out->combining_inline))
 	    {
-	      grub_errno = GRUB_ERR_NONE;
-	      continue;
+	      n = grub_realloc (out->combining_ptr,
+				sizeof (n[0]) * (out->ncomb + 1));
+	      if (!n)
+		{
+		  grub_errno = GRUB_ERR_NONE;
+		  continue;
+		}
+	      out->combining_ptr = n;
+	    }
+	  else
+	    {
+	      n = grub_malloc (sizeof (n[0]) * (out->ncomb + 1));
+	      if (!n)
+		{
+		  grub_errno = GRUB_ERR_NONE;
+		  continue;
+		}
+	      grub_memcpy (n, out->combining_inline,
+			   sizeof (out->combining_inline));
+	      out->combining_ptr = n;
 	    }
-	  out->combining = n;
 
 	  for (j = last_comb_pointer; j < out->ncomb; j++)
-	    if (is_type_after (out->combining[j].type, comb_type))
+	    if (is_type_after (n[j].type, comb_type))
 	      break;
-	  grub_memmove (out->combining + j + 1,
-			out->combining + j,
+	  grub_memmove (n + j + 1,
+			n + j,
 			(out->ncomb - j)
-			* sizeof (out->combining[0]));
-	  out->combining = n;
-	  out->combining[j].code = *ptr;
-	  out->combining[j].type = comb_type;
+			* sizeof (n[0]));
+	  n[j].code = *ptr;
+	  n[j].type = comb_type;
 	  out->ncomb++;
 	  continue;
 	}
@@ -618,7 +519,6 @@ grub_unicode_aglomerate_comb (const grub_uint32_t *in, grub_size_t inlen,
       out->attributes = 0;
       out->ncomb = 0;
       out->estimated_width = 1;
-      out->combining = NULL;
     }
   return ptr - in;
 }
@@ -626,9 +526,13 @@ grub_unicode_aglomerate_comb (const grub_uint32_t *in, grub_size_t inlen,
 static grub_ssize_t
 bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 		struct grub_unicode_glyph *visual,
-		grub_size_t visual_len, unsigned *levels,
-		grub_ssize_t (*getcharwidth) (const struct grub_unicode_glyph *visual),
-		grub_size_t maxwidth, grub_size_t startwidth)
+		grub_size_t visual_len,
+		grub_ssize_t (*getcharwidth) (const struct grub_unicode_glyph *visual, void *getcharwidth_arg),
+		void *getcharwidth_arg,
+		grub_size_t maxwidth, grub_size_t startwidth,
+		grub_uint32_t contchar,
+		struct grub_term_pos *pos, int primitive_wrap,
+		grub_size_t log_end)
 {
   struct grub_unicode_glyph *outptr = visual_out;
   unsigned line_start = 0;
@@ -636,20 +540,30 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
   unsigned k;
   grub_ssize_t last_space = -1;
   grub_ssize_t last_space_width = 0;
+  int lines = 0;
 
   auto void revert (unsigned start, unsigned end);
   void revert (unsigned start, unsigned end)
   {
     struct grub_unicode_glyph t;
-    unsigned i, tl;
-    for (i = 0; i <= (end - start) / 2; i++)
+    unsigned i;
+    int a = 0, b = 0;
+    if (pos)
+      {
+	a = pos[visual[start].orig_pos].x;
+	b = pos[visual[end].orig_pos].x;
+      }
+    for (i = 0; i < (end - start) / 2 + 1; i++)
       {
 	t = visual[start + i];
 	visual[start + i] = visual[end - i];
 	visual[end - i] = t;
-	tl = levels[start + i];
-	levels[start + i] = levels[end - i];
-	levels[end - i] = tl;
+
+	if (pos)
+	  {
+	    pos[visual[start + i].orig_pos].x = a + b - pos[visual[start + i].orig_pos].x;
+	    pos[visual[end - i].orig_pos].x = a + b - pos[visual[end - i].orig_pos].x;
+	  }
       }
   }
 
@@ -660,10 +574,27 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
     {
       grub_ssize_t last_width = 0;
 
-      if (getcharwidth && k != visual_len)
-	line_width += last_width = getcharwidth (&visual[k]);
+ 
+      if (pos && k != visual_len)
+	{
+	  pos[visual[k].orig_pos].x = line_width;
+	  pos[visual[k].orig_pos].y = lines;
+	  pos[visual[k].orig_pos].valid = 1;
+	}
 
-      if (k != visual_len && visual[k].base == ' ')
+      if (k == visual_len && pos)
+	{
+	  pos[log_end].x = line_width;
+	  pos[log_end].y = lines;
+	  pos[log_end].valid = 1;
+	}
+
+      if (getcharwidth && k != visual_len)
+	line_width += last_width = getcharwidth (&visual[k], getcharwidth_arg);
+
+      if (k != visual_len && (visual[k].base == ' '
+			      || visual[k].base == '\t')
+	  && !primitive_wrap)
 	{
 	  last_space = k;
 	  last_space_width = line_width;
@@ -675,9 +606,12 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	  unsigned min_odd_level = 0xffffffff;
 	  unsigned max_level = 0;
 
+	  lines++;
+
 	  if (k != visual_len && last_space > (signed) line_start)
 	    k = last_space;
-	  else if (k != visual_len && line_start == 0 && startwidth != 0)
+	  else if (k != visual_len && line_start == 0 && startwidth != 0
+		   && !primitive_wrap)
 	    {
 	      k = 0;
 	      last_space_width = startwidth;
@@ -689,25 +623,27 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	    unsigned i;
 	    for (i = line_start; i < k; i++)
 	      {
-		if (levels[i] > max_level)
-		  max_level = levels[i];
-		if (levels[i] < min_odd_level && (levels[i] & 1))
-		  min_odd_level = levels[i];
+		if (visual[i].bidi_level > max_level)
+		  max_level = visual[i].bidi_level;
+		if (visual[i].bidi_level < min_odd_level && (visual[i].bidi_level & 1))
+		  min_odd_level = visual[i].bidi_level;
 	      }
 	  }
 
 	  {
 	    unsigned j;	  
 	    /* FIXME: can be optimized.  */
-	    for (j = max_level; j >= min_odd_level; j--)
+	    for (j = max_level; j > min_odd_level - 1; j--)
 	      {
-		unsigned in = 0;
+		unsigned in = line_start;
 		unsigned i;
 		for (i = line_start; i < k; i++)
 		  {
-		    if (i != line_start && levels[i] >= j && levels[i-1] < j)
+		    if (i != line_start && visual[i].bidi_level >= j
+			&& visual[i-1].bidi_level < j)
 		      in = i;
-		    if (levels[i] >= j && (i + 1 == k || levels[i+1] < j))
+		    if (visual[i].bidi_level >= j && (i + 1 == k
+						 || visual[i+1].bidi_level < j))
 		      revert (in, i);
 		  }
 	      }
@@ -717,10 +653,10 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	    unsigned i;
 	    for (i = line_start; i < k; i++)
 	      {
-		if (is_mirrored (visual[i].base) && levels[i])
+		if (is_mirrored (visual[i].base) && visual[i].bidi_level)
 		  visual[i].attributes |= GRUB_UNICODE_GLYPH_ATTRIBUTE_MIRROR;
 		if ((visual[i].attributes & GRUB_UNICODE_GLYPH_ATTRIBUTES_JOIN)
-		    && levels[i])
+		    && visual[i].bidi_level)
 		  {
 		    int left, right;
 		    left = visual[i].attributes
@@ -768,7 +704,8 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	  {
 	    int right_join = 0;
 	    signed i;
-	    for (i = k - 1; i >= (signed) line_start; i--)
+	    for (i = k - 1; i >= 0 && (unsigned) i + 1 > line_start;
+		 i--)
 	      {
 		enum grub_join_type join_type = get_join_type (visual[i].base);
 		if (!(visual[i].attributes
@@ -798,6 +735,12 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	  outptr += k - line_start;
 	  if (k != visual_len)
 	    {
+	      if (contchar)
+		{
+		  grub_memset (outptr, 0, sizeof (visual[0]));
+		  outptr->base = contchar;
+		  outptr++;
+		}
 	      grub_memset (outptr, 0, sizeof (visual[0]));
 	      outptr->base = '\n';
 	      outptr++;
@@ -808,6 +751,11 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 
 	  line_start = k;
 	  line_width -= last_space_width;
+	  if (pos && k != visual_len)
+	    {
+	      pos[visual[k].orig_pos].x = 0;
+	      pos[visual[k].orig_pos].y = lines;
+	    }
 	}
     }
 
@@ -819,13 +767,16 @@ static grub_ssize_t
 grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 				  grub_size_t logical_len,
 				  struct grub_unicode_glyph *visual_out,
-				  grub_ssize_t (*getcharwidth) (const struct grub_unicode_glyph *visual),
-				  grub_size_t maxwidth, grub_size_t startwidth)
+				  grub_ssize_t (*getcharwidth) (const struct grub_unicode_glyph *visual, void *getcharwidth_arg),
+				  void *getcharwidth_arg,
+				  grub_size_t maxwidth, grub_size_t startwidth,
+				  grub_uint32_t contchar,
+				  struct grub_term_pos *pos,
+				  int primitive_wrap,
+				  grub_size_t log_end)
 {
   enum grub_bidi_type type = GRUB_BIDI_TYPE_L;
   enum override_status {OVERRIDE_NEUTRAL = 0, OVERRIDE_R, OVERRIDE_L};
-  unsigned *levels;
-  enum grub_bidi_type *resolved_types;
   unsigned base_level;
   enum override_status cur_override;
   unsigned i;
@@ -869,24 +820,9 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
     cur_override = stack_override[stack_depth];
   }
 
-  levels = grub_malloc (sizeof (levels[0]) * logical_len);
-  if (!levels)
-    return -1;
-
-  resolved_types = grub_malloc (sizeof (resolved_types[0]) * logical_len);
-  if (!resolved_types)
-    {
-      grub_free (levels);
-      return -1;
-    }
-
   visual = grub_malloc (sizeof (visual[0]) * logical_len);
   if (!visual)
-    {
-      grub_free (resolved_types);
-      grub_free (levels);
-      return -1;
-    }
+    return -1;
 
   for (i = 0; i < logical_len; i++)
     {
@@ -939,9 +875,13 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 	    continue;
 	  }
 
+	/* The tags: deprecated, never used.  */
+	if (*lptr >= GRUB_UNICODE_TAG_START && *lptr <= GRUB_UNICODE_TAG_END)
+	  continue;
+
 	p = grub_unicode_aglomerate_comb (lptr, logical + logical_len - lptr, 
 					  &visual[visual_len]);
-	
+	visual[visual_len].orig_pos = lptr - logical;
 	type = get_bidi_type (visual[visual_len].base);
 	switch (type)
 	  {
@@ -987,13 +927,13 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 	      join_state = JOIN_DEFAULT;
 	      zwj_propagate_to_previous = 1;
 
-	      levels[visual_len] = cur_level;
+	      visual[visual_len].bidi_level = cur_level;
 	      if (cur_override != OVERRIDE_NEUTRAL)
-		resolved_types[visual_len] = 
+		visual[visual_len].bidi_type = 
 		  (cur_override == OVERRIDE_L) ? GRUB_BIDI_TYPE_L
 		  : GRUB_BIDI_TYPE_R;
 	      else
-		resolved_types[visual_len] = type;
+		visual[visual_len].bidi_type = type;
 	      visual_len++;
 	    }
 	  }
@@ -1008,16 +948,16 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 	  unsigned prev_level, next_level, cur_run_level;
 	  unsigned last_type, last_strong_type;
 	  for (run_end = run_start; run_end < visual_len &&
-		 levels[run_end] == levels[run_start]; run_end++);
+		 visual[run_end].bidi_level == visual[run_start].bidi_level; run_end++);
 	  if (run_start == 0)
 	    prev_level = base_level;
 	  else
-	    prev_level = levels[run_start - 1];
+	    prev_level = visual[run_start - 1].bidi_level;
 	  if (run_end == visual_len)
 	    next_level = base_level;
 	  else
-	    next_level = levels[run_end];
-	  cur_run_level = levels[run_start];
+	    next_level = visual[run_end].bidi_level;
+	  cur_run_level = visual[run_start].bidi_level;
 	  if (prev_level & 1)
 	    last_type = GRUB_BIDI_TYPE_R;
 	  else
@@ -1025,48 +965,48 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 	  last_strong_type = last_type;
 	  for (i = run_start; i < run_end; i++)
 	    {
-	      switch (resolved_types[i])
+	      switch (visual[i].bidi_type)
 		{
 		case GRUB_BIDI_TYPE_NSM:
-		  resolved_types[i] = last_type;
+		  visual[i].bidi_type = last_type;
 		  break;
 		case GRUB_BIDI_TYPE_EN:
 		  if (last_strong_type == GRUB_BIDI_TYPE_AL)
-		    resolved_types[i] = GRUB_BIDI_TYPE_AN;
+		    visual[i].bidi_type = GRUB_BIDI_TYPE_AN;
 		  break;
 		case GRUB_BIDI_TYPE_L:
 		case GRUB_BIDI_TYPE_R:
-		  last_strong_type = resolved_types[i];
+		  last_strong_type = visual[i].bidi_type;
 		  break;
 		case GRUB_BIDI_TYPE_ES:
 		  if (last_type == GRUB_BIDI_TYPE_EN
 		      && i + 1 < run_end 
-		      && resolved_types[i + 1] == GRUB_BIDI_TYPE_EN)
-		    resolved_types[i] = GRUB_BIDI_TYPE_EN;
+		      && visual[i + 1].bidi_type == GRUB_BIDI_TYPE_EN)
+		    visual[i].bidi_type = GRUB_BIDI_TYPE_EN;
 		  else
-		    resolved_types[i] = GRUB_BIDI_TYPE_ON;
+		    visual[i].bidi_type = GRUB_BIDI_TYPE_ON;
 		  break;
 		case GRUB_BIDI_TYPE_ET:
 		  {
 		    unsigned j;
 		    if (last_type == GRUB_BIDI_TYPE_EN)
 		      {
-			resolved_types[i] = GRUB_BIDI_TYPE_EN;
+			visual[i].bidi_type = GRUB_BIDI_TYPE_EN;
 			break;
 		      }
 		    for (j = i; j < run_end
-			   && resolved_types[j] == GRUB_BIDI_TYPE_ET; j++);
-		    if (j != run_end && resolved_types[j] == GRUB_BIDI_TYPE_EN)
+			   && visual[j].bidi_type == GRUB_BIDI_TYPE_ET; j++);
+		    if (j != run_end && visual[j].bidi_type == GRUB_BIDI_TYPE_EN)
 		      {
 			for (; i < run_end
-			       && resolved_types[i] == GRUB_BIDI_TYPE_ET; i++)
-			  resolved_types[i] = GRUB_BIDI_TYPE_EN;
+			       && visual[i].bidi_type == GRUB_BIDI_TYPE_ET; i++)
+			  visual[i].bidi_type = GRUB_BIDI_TYPE_EN;
 			i--;
 			break;
 		      }
 		    for (; i < run_end
-			   && resolved_types[i] == GRUB_BIDI_TYPE_ET; i++)
-		      resolved_types[i] = GRUB_BIDI_TYPE_ON;
+			   && visual[i].bidi_type == GRUB_BIDI_TYPE_ET; i++)
+		      visual[i].bidi_type = GRUB_BIDI_TYPE_ON;
 		    i--;
 		    break;		
 		  }
@@ -1074,33 +1014,33 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 		case GRUB_BIDI_TYPE_CS:
 		  if (last_type == GRUB_BIDI_TYPE_EN
 		      && i + 1 < run_end 
-		      && resolved_types[i + 1] == GRUB_BIDI_TYPE_EN)
+		      && visual[i + 1].bidi_type == GRUB_BIDI_TYPE_EN)
 		    {
-		      resolved_types[i] = GRUB_BIDI_TYPE_EN;
+		      visual[i].bidi_type = GRUB_BIDI_TYPE_EN;
 		      break;
 		    }
 		  if (last_type == GRUB_BIDI_TYPE_AN
 		      && i + 1 < run_end 
-		      && (resolved_types[i + 1] == GRUB_BIDI_TYPE_AN
-			  || (resolved_types[i + 1] == GRUB_BIDI_TYPE_EN
+		      && (visual[i + 1].bidi_type == GRUB_BIDI_TYPE_AN
+			  || (visual[i + 1].bidi_type == GRUB_BIDI_TYPE_EN
 			      && last_strong_type == GRUB_BIDI_TYPE_AL)))
 		    {
-		      resolved_types[i] = GRUB_BIDI_TYPE_EN;
+		      visual[i].bidi_type = GRUB_BIDI_TYPE_EN;
 		      break;
 		    }
-		  resolved_types[i] = GRUB_BIDI_TYPE_ON;
+		  visual[i].bidi_type = GRUB_BIDI_TYPE_ON;
 		  break;
 		case GRUB_BIDI_TYPE_AL:
-		  last_strong_type = resolved_types[i];
-		  resolved_types[i] = GRUB_BIDI_TYPE_R;
+		  last_strong_type = visual[i].bidi_type;
+		  visual[i].bidi_type = GRUB_BIDI_TYPE_R;
 		  break;
 		default: /* Make GCC happy.  */
 		  break;
 		}
-	      last_type = resolved_types[i];
-	      if (resolved_types[i] == GRUB_BIDI_TYPE_EN
+	      last_type = visual[i].bidi_type;
+	      if (visual[i].bidi_type == GRUB_BIDI_TYPE_EN
 		  && last_strong_type == GRUB_BIDI_TYPE_L)
-		resolved_types[i] = GRUB_BIDI_TYPE_L;
+		visual[i].bidi_type = GRUB_BIDI_TYPE_L;
 	    }
 	  if (prev_level & 1)
 	    last_type = GRUB_BIDI_TYPE_R;
@@ -1111,13 +1051,13 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 	      unsigned j;
 	      unsigned next_type;
 	      for (j = i; j < run_end &&
-		     (resolved_types[j] == GRUB_BIDI_TYPE_B
-		      || resolved_types[j] == GRUB_BIDI_TYPE_S
-		      || resolved_types[j] == GRUB_BIDI_TYPE_WS
-		      || resolved_types[j] == GRUB_BIDI_TYPE_ON); j++);
+		     (visual[j].bidi_type == GRUB_BIDI_TYPE_B
+		      || visual[j].bidi_type == GRUB_BIDI_TYPE_S
+		      || visual[j].bidi_type == GRUB_BIDI_TYPE_WS
+		      || visual[j].bidi_type == GRUB_BIDI_TYPE_ON); j++);
 	      if (j == i)
 		{
-		  if (resolved_types[i] == GRUB_BIDI_TYPE_L)
+		  if (visual[i].bidi_type == GRUB_BIDI_TYPE_L)
 		    last_type = GRUB_BIDI_TYPE_L;
 		  else
 		    last_type = GRUB_BIDI_TYPE_R;
@@ -1128,39 +1068,39 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 		next_type = (next_level & 1) ? GRUB_BIDI_TYPE_R : GRUB_BIDI_TYPE_L;
 	      else
 		{
-		  if (resolved_types[j] == GRUB_BIDI_TYPE_L)
+		  if (visual[j].bidi_type == GRUB_BIDI_TYPE_L)
 		    next_type = GRUB_BIDI_TYPE_L;
 		  else
 		    next_type = GRUB_BIDI_TYPE_R;
 		}
 	      if (next_type == last_type)
 		for (; i < j; i++)
-		  resolved_types[i] = last_type;
+		  visual[i].bidi_type = last_type;
 	      else
 		for (; i < j; i++)
-		  resolved_types[i] = (cur_run_level & 1) ? GRUB_BIDI_TYPE_R
+		  visual[i].bidi_type = (cur_run_level & 1) ? GRUB_BIDI_TYPE_R
 		    : GRUB_BIDI_TYPE_L;
 	    }
 	}
 
       for (i = 0; i < visual_len; i++)
 	{
-	  if (!(levels[i] & 1) && resolved_types[i] == GRUB_BIDI_TYPE_R)
+	  if (!(visual[i].bidi_level & 1) && visual[i].bidi_type == GRUB_BIDI_TYPE_R)
 	    {
-	      levels[i]++;
+	      visual[i].bidi_level++;
 	      continue;
 	    }
-	  if (!(levels[i] & 1) && (resolved_types[i] == GRUB_BIDI_TYPE_AN
-				   || resolved_types[i] == GRUB_BIDI_TYPE_EN))
+	  if (!(visual[i].bidi_level & 1) && (visual[i].bidi_type == GRUB_BIDI_TYPE_AN
+				   || visual[i].bidi_type == GRUB_BIDI_TYPE_EN))
 	    {
-	      levels[i] += 2;
+	      visual[i].bidi_level += 2;
 	      continue;
 	    }
-	  if ((levels[i] & 1) && (resolved_types[i] == GRUB_BIDI_TYPE_L
-				  || resolved_types[i] == GRUB_BIDI_TYPE_AN
-				  || resolved_types[i] == GRUB_BIDI_TYPE_EN))
+	  if ((visual[i].bidi_level & 1) && (visual[i].bidi_type == GRUB_BIDI_TYPE_L
+				  || visual[i].bidi_type == GRUB_BIDI_TYPE_AN
+				  || visual[i].bidi_type == GRUB_BIDI_TYPE_EN))
 	    {
-	      levels[i]++;
+	      visual[i].bidi_level++;
 	      continue;
 	    }
 	}
@@ -1168,15 +1108,14 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
   else
     {
       for (i = 0; i < visual_len; i++)
-	levels[i] = 0;
+	visual[i].bidi_level = 0;
     }
-  grub_free (resolved_types);
 
   {
     grub_ssize_t ret;
-    ret = bidi_line_wrap (visual_out, visual, visual_len, levels, 
-			  getcharwidth, maxwidth, startwidth);
-    grub_free (levels);
+    ret = bidi_line_wrap (visual_out, visual, visual_len,
+			  getcharwidth, getcharwidth_arg, maxwidth, startwidth, contchar,
+			  pos, primitive_wrap, log_end);
     grub_free (visual);
     return ret;
   }
@@ -1186,12 +1125,14 @@ grub_ssize_t
 grub_bidi_logical_to_visual (const grub_uint32_t *logical,
 			     grub_size_t logical_len,
 			     struct grub_unicode_glyph **visual_out,
-			     grub_ssize_t (*getcharwidth) (const struct grub_unicode_glyph *visual),
-			     grub_size_t max_length, grub_size_t startwidth)
+			     grub_ssize_t (*getcharwidth) (const struct grub_unicode_glyph *visual, void *getcharwidth_arg),
+			     void *getcharwidth_arg,
+			     grub_size_t max_length, grub_size_t startwidth,
+			     grub_uint32_t contchar, struct grub_term_pos *pos, int primitive_wrap)
 {
   const grub_uint32_t *line_start = logical, *ptr;
   struct grub_unicode_glyph *visual_ptr;
-  *visual_out = visual_ptr = grub_malloc (2 * sizeof (visual_ptr[0])
+  *visual_out = visual_ptr = grub_malloc (3 * sizeof (visual_ptr[0])
 					  * logical_len);
   if (!visual_ptr)
     return -1;
@@ -1204,8 +1145,13 @@ grub_bidi_logical_to_visual (const grub_uint32_t *logical,
 						  ptr - line_start,
 						  visual_ptr,
 						  getcharwidth,
+						  getcharwidth_arg,
 						  max_length,
-						  startwidth);
+						  startwidth,
+						  contchar,
+						  pos,
+						  primitive_wrap,
+						  logical_len);
 	  startwidth = 0;
 
 	  if (ret < 0)
@@ -1272,3 +1218,52 @@ grub_unicode_shape_code (grub_uint32_t in, grub_uint8_t attr)
 
   return in;
 }
+
+const grub_uint32_t *
+grub_unicode_get_comb_start (const grub_uint32_t *str, 
+			     const grub_uint32_t *cur)
+{
+  const grub_uint32_t *ptr;
+  for (ptr = cur; ptr >= str; ptr--)
+    {
+      if (*ptr >= GRUB_UNICODE_VARIATION_SELECTOR_1
+	  && *ptr <= GRUB_UNICODE_VARIATION_SELECTOR_16)
+	continue;
+
+      if (*ptr >= GRUB_UNICODE_VARIATION_SELECTOR_17
+	  && *ptr <= GRUB_UNICODE_VARIATION_SELECTOR_256)
+	continue;
+	
+      enum grub_comb_type comb_type;
+      comb_type = grub_unicode_get_comb_type (*ptr);
+      if (comb_type)
+	continue;
+      return ptr;
+    }
+  return str;
+}
+
+const grub_uint32_t *
+grub_unicode_get_comb_end (const grub_uint32_t *end, 
+			   const grub_uint32_t *cur)
+{
+  const grub_uint32_t *ptr;
+  for (ptr = cur; ptr < end; ptr++)
+    {
+      if (*ptr >= GRUB_UNICODE_VARIATION_SELECTOR_1
+	  && *ptr <= GRUB_UNICODE_VARIATION_SELECTOR_16)
+	continue;
+
+      if (*ptr >= GRUB_UNICODE_VARIATION_SELECTOR_17
+	  && *ptr <= GRUB_UNICODE_VARIATION_SELECTOR_256)
+	continue;
+	
+      enum grub_comb_type comb_type;
+      comb_type = grub_unicode_get_comb_type (*ptr);
+      if (comb_type)
+	continue;
+      return ptr;
+    }
+  return end;
+}
+
