@@ -1,7 +1,7 @@
 /* lvm.c - LVM support for GRUB utils.  */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2006,2007,2008  Free Software Foundation, Inc.
+ *  Copyright (C) 2006,2007,2008,2011  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -17,8 +17,7 @@
  *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-/* We only support LVM on Linux.  */
-#ifdef __linux__
+#if defined(__linux__) || defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
 #include <grub/emu/misc.h>
 #include <grub/util/misc.h>
 #include <grub/util/lvm.h>
@@ -26,22 +25,20 @@
 #include <string.h>
 #include <sys/stat.h>
 
-#define LVM_DEV_MAPPER_STRING "/dev/mapper/"
-
 int
 grub_util_lvm_isvolume (char *name)
 {
-  char *devname;
+  char *lvmdevname;
   struct stat st;
   int err;
 
-  devname = xmalloc (strlen (name) + sizeof (LVM_DEV_MAPPER_STRING));
+  lvmdevname = xmalloc (strlen (name) + sizeof (LVM_DEV_MAPPER_STRING));
 
-  strcpy (devname, LVM_DEV_MAPPER_STRING);
-  strcpy (devname + sizeof(LVM_DEV_MAPPER_STRING) - 1, name);
+  strcpy (lvmdevname, LVM_DEV_MAPPER_STRING);
+  strcpy (lvmdevname + sizeof(LVM_DEV_MAPPER_STRING) - 1, name);
 
-  err = stat (devname, &st);
-  free (devname);
+  err = stat (lvmdevname, &st);
+  free (lvmdevname);
 
   if (err)
     return 0;
@@ -49,4 +46,4 @@ grub_util_lvm_isvolume (char *name)
     return 1;
 }
 
-#endif /* ! __linux__ */
+#endif
